{*******************************************************}
{                                                       }
{       RichView                                        }
{       TRichViewTarget - component accepting drag&drop }
{                                                       }
{       Copyright (c) Sergey Tkachenko                  }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{                                                       }
{*******************************************************}

unit MultiDropTarget;

{$I RV_Defs.inc}

interface

uses Windows, SysUtils, Messages, ActiveX, Classes, Controls, Graphics, RVStyle, RichView,
  RVDragDrop, RVStr, RVUni, ShlObj, RVFuncs, RVEdit, RVClasses, RVTypes, RVGrHandler;

type
  TRVAfterDropEvent = procedure (Sender: TCustomRichView; X, Y: Integer;
    Format: TRVDragDropFormat; const Title: String) of object;
  TRVDragEvent = procedure (Sender: TCustomRichView; X, Y: Integer;
    var CanDropHere: Boolean) of object;

  { TRichViewTarget accepts drag&drop and loads its result in Receiver property }
  { This component is not registered, so it can be created at runtime only.
    You can add Register procedure, if you want }
  TRichViewTarget = class (TRichView)
  private
    FDropTarget: TRVDropTarget;
    FOnOleDragEnter, FOnOleDragLeave, FOnOleDropNotAccepted: TNotifyEvent;
    FOnOleAfterDrop: TRVAfterDropEvent;
    FOnOleDragOver, FOnOleBeforeDrop: TRVDragEvent;
    FReceiver: TCustomRichView;
    procedure WMCreate(var Message: TMessage); message WM_CREATE;
    procedure WMDestroy(var Message: TMessage); message WM_DESTROY;
  protected
    function OleDragEnter(X,Y: Integer): Boolean; override;
    procedure CallOleDragEnterEvent(const DataObj: IDataObject;
      KeyState: Integer; pt: TPoint; PossibleEffects: Integer;
      var Effect: Integer); override;
    procedure OleDragLeave; override;
    function OleDragOver(X, Y: Integer): Boolean; override;
    procedure CallOleDragOverEvent(KeyState: Integer; pt: TPoint;
      PossibleEffects: Integer; var Effect: Integer); override;
    function OleDrop(const DataObj: IDataObject; FMove: Boolean;
      KeyState: Integer; pt: TPoint; PossibleEffects: Integer): Integer; override;
    procedure ReleaseOleDropTargetObject; override;
    function OleCanAcceptFormat(Format: Word): Boolean; override;
  published
    property OnOleDragEnter: TNotifyEvent read FOnOleDragEnter write FOnOleDragEnter;
    property OnOleDragLeave: TNotifyEvent read FOnOleDragLeave write FOnOleDragLeave;
    property OnOleDragOver: TRVDragEvent read FOnOleDragOver write FOnOleDragOver;
    property OnOleBeforeDrop: TRVDragEvent read FOnOleBeforeDrop write FOnOleBeforeDrop;
    property OnOleDropNotAccepted: TNotifyEvent read FOnOleDropNotAccepted write FOnOleDropNotAccepted;
    property OnOleAfterDrop: TRVAfterDropEvent read FOnOleAfterDrop write FOnOleAfterDrop;
    property Receiver: TCustomRichView read FReceiver write FReceiver;
  end;

implementation

{ TRichViewTarget }

procedure TRichViewTarget.CallOleDragEnterEvent(const DataObj: IDataObject;
  KeyState: Integer; pt: TPoint; PossibleEffects: Integer;
  var Effect: Integer);
begin
  if Assigned(FOnOleDragEnter) then
    FOnOleDragEnter(Self);
end;

procedure TRichViewTarget.CallOleDragOverEvent(KeyState: Integer;
  pt: TPoint; PossibleEffects: Integer; var Effect: Integer);
var CanDropHere: Boolean;
begin
  CanDropHere := True;
  if Assigned(FOnOleDragOver) then
    FOnOleDragOver(Self, pt.X, pt.Y, CanDropHere);
  if CanDropHere then begin
    Effect := DROPEFFECT_COPY;
    if (Effect and PossibleEffects) = 0 then
      Effect := DROPEFFECT_LINK
    end
  else
    Effect := DROPEFFECT_NONE;
end;

function TRichViewTarget.OleCanAcceptFormat(Format: Word): Boolean;
begin
  Result := True;
end;

function TRichViewTarget.OleDragEnter(X, Y: Integer): Boolean;
begin
  Result := True;
end;

procedure TRichViewTarget.OleDragLeave;
begin
  if Assigned(FOnOleDragLeave) then
    FOnOleDragLeave(Self);
end;

function TRichViewTarget.OleDragOver(X, Y: Integer): Boolean;
begin
  Result := True;
end;

// Converting text from String to internal representation
function ConvertStringToItemText(const Text: String;
  UnicodeItem: Boolean; CodePage: Cardinal): TRVRawByteString;
begin
  {$IFDEF RVUNICODESTR} // <-- declared in RV_Defs.inc
  if UnicodeItem then
    Result := RVU_GetRawUnicode(Text)
  else
    Result := TRVAnsiString(Text);
  {$ELSE}
  if UnicodeItem then
    Result := RVU_AnsiToUnicode(CodePage, Text)
  else
    Result := Text;
  {$ENDIF}
end;

// Converting text from internal representation to String
function ConvertItemTextToString(const ItemText: TRVRawByteString;
  UnicodeItem: Boolean; CodePage: Cardinal): String;
begin
  {$IFDEF RVUNICODESTR} // <-- declared in RV_Defs.inc
  // Delphi 2009+: String is Unicode
  if UnicodeItem then
    Result := RVU_RawUnicodeToWideString(ItemText)
  else
    Result := RVU_RawUnicodeToWideString(
      RVU_AnsiToUnicode(CodePage, ItemText));
  {$ELSE}
  // Delphi 4-2007: String is ANSI
  if UnicodeItem then
    Result := RVU_UnicodeToAnsi(CodePage, ItemText)
  else
    Result := ItemText;
  {$ENDIF}
end;

function TRichViewTarget.OleDrop(const DataObj: IDataObject;
  FMove: Boolean; KeyState: Integer; pt: TPoint;
  PossibleEffects: Integer): Integer;
var Stream: TRVMemoryStream;
    Title, Target, DropTitle: TRVUnicodeString;
    Str: TRVUnicodeString;
    StrA: TRVAnsiString;
    cf: Word;
    i, StyleNo: Integer;
    Tag: TRVTag;
    StgMedium: TStgMedium;
    PFGD: PFileGroupDescriptor;
    bmp: TGraphic;
    wmf: TGraphic;
    Files: TStringList;
    DoDefault: Boolean;
    FileAction: TRVDropFileAction;
    Format: TRVDragDropFormat;
    FileAccepted, CanDropHere: Boolean;

    procedure Complete;
    begin
      FReceiver.Format;
      if Assigned(FOnOleAfterDrop) then
        FOnOleAfterDrop(Self, pt.X, pt.Y, Format, DropTitle);
    end;

    function AddFile(const FileName: String): Boolean;
    var pic: TPicture;
        gr: TGraphic;
        Ext: String;
    begin
      Result := False;
      try
        pic := TPicture.Create;
        try
          pic.LoadFromFile(FileName);
          gr := RVGraphicHandler.CreateGraphic(TGraphicClass(pic.Graphic.ClassType));
          gr.Assign(pic.Graphic);
          FReceiver.AddPicture( '', gr, 0, rvvaBaseline);
          Result := True;
        except;
        end;
        pic.Free;
        if Result then
          exit;
        Ext := LowerCase(ExtractFileExt(FileName));
        if Ext='.rtf' then
        begin
          FReceiver.LoadRTF(FileName);
          Result := True;
          exit;
        end;
        if Ext='.rvf' then
        begin
          FReceiver.LoadRVF(FileName);
          Result := True;
          exit;
        end;
        if (Ext='.txt') or (Ext='.pas') or (Ext='.cpp') or
           (Ext='.hpp') or (Ext='.h') or (Ext='.ini') or (Ext='.xml') or
           (Ext='.html') or (Ext='.htm') or (Ext='.xhtml') then
        begin
          if RV_TestFileUnicode(FileName)=rvutYes then
            FReceiver.LoadTextW(FileName, 0, 0, False)
          else
            FReceiver.LoadText(FileName, 0, 0, False);
          Result := True;
          exit;
        end;
      except
      end;
    end;

begin
  CanDropHere := True;
  if Assigned(FOnOleBeforeDrop) then
    FOnOleBeforeDrop(Self, pt.X, pt.Y, CanDropHere);
  if (FReceiver = nil) or not CanDropHere then
  begin
    Result := DROPEFFECT_NONE;
    exit;
  end;
  FReceiver.Clear;
  Result := DROPEFFECT_COPY;
  if FDropTarget.HasFormat(DataObj, CFRV_RVF) then
  begin
    Stream := FDropTarget.GetAsStream(DataObj, CFRV_RVF);
    if Stream<>nil then
    begin
      FReceiver.LoadRVFFromStream(Stream);
      Stream.Free;
      DropTitle := 'RichView Format';
      Format := rvddRVF;
      Complete;
      exit;
    end;
  end;
  if FDropTarget.HasFormat(DataObj, CFRV_RTF) then
  begin
    Stream := FDropTarget.GetAsStream(DataObj, CFRV_RTF);
    if Stream<>nil then begin
      FReceiver.LoadRTFFromStream(Stream);
      Stream.Free;
      DropTitle := 'Rich Text Format';
      Format := rvddRTF;      
      Complete;
      exit;
    end;
  end;
  if FDropTarget.HasFormat(DataObj, CFRV_URL) and
    FDropTarget.GetAsTextA(DataObj, CFRV_URL, StrA) then
  begin
    Target := RVU_AnsiToUnicode(CP_ACP, StrA);
    Title := Target;
    cf := RegisterClipboardFormat(CFSTR_FILEDESCRIPTOR);
    if FDropTarget.HasFormat(DataObj, cf) and
      FDropTarget.GetMedium(DataObj, cf, StgMedium) then
    begin
      PFGD := Pointer(GlobalLock(StgMedium.HGlobal));
      Title := pFGD^.fgd[0].cFileName;
      GlobalUnlock(StgMedium.HGlobal);
      if Length(Title)>4 then
        Delete(Title, Length(Title)-3,4)
      else
        Title := Target;
      ReleaseStgMedium(StgMedium);
    end;
    StyleNo := 0;
    Tag := '';
    if Assigned(FReceiver.OnReadHyperlink) then
      FReceiver.OnReadHyperlink(FReceiver, Target, '', rvlfURL, StyleNo, Tag, Title);
    FReceiver.AddNL(Title, StyleNo, 0, Tag);
    DropTitle := 'Hyperlink';
    Format := rvddURL;
    Complete;
    exit;
  end;
  if FDropTarget.HasFormat(DataObj, CF_UNICODETEXT) and
    FDropTarget.GetAsTextW(DataObj, CF_UNICODETEXT, Str) then
  begin
    FReceiver.AddTextNLW(Str, 0, 0, 0, False);
    DropTitle := 'Unicode Text';
    Format := rvddUnicodeText;
    Complete;
    exit;
  end;
  if FDropTarget.HasFormat(DataObj, CF_TEXT) and
    FDropTarget.GetAsTextA(DataObj, CF_TEXT, StrA) then
  begin
    FReceiver.AddTextNLA(StrA, 0, 0, 0);
    DropTitle := 'Text';
    Format := rvddText;    
    Complete;
    exit;
  end;
  if FDropTarget.HasFormat(DataObj, CF_BITMAP) or
    FDropTarget.HasFormat(DataObj, CF_DIB) then
  begin
    bmp := FDropTarget.GetAsBitmap(DataObj, True);
    if (bmp<>nil) and not ((bmp is TBitmap) and bmp.Empty) then
    begin
      FReceiver.AddPicture('', bmp, 0, rvvaBaseline);
      DropTitle := 'Bitmap';
      Format := rvddBitmap;
      Complete;
      exit;
    end;
    bmp.Free;
  end;
  if FDropTarget.HasFormat(DataObj, CF_ENHMETAFILE) then
  begin
    wmf := FDropTarget.GetAsMetafile(DataObj);
    if (wmf<>nil) then
    begin
      FReceiver.AddPicture('', wmf, 0, rvvaBaseline);
      DropTitle := 'Metafile';
      Format := rvddMetafile;            
      Complete;
      exit;
    end;
  end;
  if FDropTarget.HasFormat(DataObj, CF_HDROP) then
  begin
    Files := FDropTarget.GetAsFiles(DataObj);
    if (Files<>nil) then begin
      DoDefault := True;
      FileAction := rvdfNone;
      if (FReceiver is TCustomRichViewEdit) and
        Assigned(TCustomRichViewEdit(FReceiver).OnDropFiles) then
        TCustomRichViewEdit(FReceiver).OnDropFiles(TCustomRichViewEdit(FReceiver),
          Files, FileAction, DoDefault);
      FileAccepted := True;
      if DoDefault then
        FileAccepted := False;
        for i := 0 to Files.Count-1 do
        begin
          FileAccepted := False;
          if AddFile(Files[i]) then
          begin
            DropTitle := Files[i];
            Format := rvddFiles;
            Complete;
            FileAccepted := True;
            if Assigned(FOnOleBeforeDrop) and (i<>Files.Count-1) then
            begin
              FOnOleBeforeDrop(Self, pt.X, pt.Y, CanDropHere);
              if (FReceiver=nil) or not CanDropHere then
                break;
              FReceiver.Clear;
              FileAccepted := False;
            end;
          end;
        end;
      Files.Free;
      if not FileAccepted and Assigned(FOnOleDropNotAccepted) then
        FOnOleDropNotAccepted(Self);
      exit;
    end;
  end;
  Result := DROPEFFECT_NONE;
  FReceiver.Format;
  if Assigned(FOnOleDropNotAccepted) then
    FOnOleDropNotAccepted(Self);
end;

procedure TRichViewTarget.ReleaseOleDropTargetObject;
begin
  FDropTarget := nil;
end;

procedure TRichViewTarget.WMCreate(var Message: TMessage);
begin
  inherited;
  if not (csDesigning in ComponentState) then begin
    if FDropTarget=nil then
      FDropTarget := TRVDropTarget.Create(Self);
    FDropTarget.RegisterDragDropWindow;
  end;
end;

procedure TRichViewTarget.WMDestroy(var Message: TMessage);
begin
  if not (csDesigning in ComponentState) and Assigned(FDropTarget) then
    FDropTarget.UnRegisterDragDropWindow;
  inherited;
end;

end.
